;----------------------------------------------------------------
; BMP chi-square attack v0.1 
; Freeware, Open Source, GPL, Copyleft, whatever you want.
;----------------------------------------------------------------
;
; Steganography statistical attack by using the chi-square test.
;
; This is my free implementation of the attack invented and
; published by Andreas Westfeld'd research group:
;
; A. Westfeld and A. Pfitzmann, "Attack on Steganographic Systems",
; Lectures Notes in Computer Science, vol. 1768, Springer-Verlag, 
; Berlin, 2000, pp. 61-75
;
; This program asks you to choose a 24-bits uncompressed 
; BMP file. It checks if the file looks like a BMP, and then
; does some calculations. It outputs two different curves along
; a grid where every vertical blue bar represents 1 ko of putative
; hidden data.
;
; The outputs are :
;
;  - in red, chi-square test to check the probability that pairs of 
;    values follow a random distribution (from start of image to 
;    successively bigger amounts of pixels). If 1, the PoVs are 
;    random, so maybe there is an hidden encrypted message in the 
;    LSBs. Or your image naturally has a random distribution of LSBs
;    (a false positive), but it is supposed to be not frequent. If 0,
;    the PoVs are not random, which is frequent in a normal image, 
;    and so there is no message hidden in the LSBs. 
;
;  - in green, the average value of all LSBs in one block of pixels 
;    (128 by default). In most natural photographs I checked, this is 
;    around 0.5 (a LSB being a bit, it can be 0 or 1). But it's 
;    clearly not the case in simple drawings. 
;
; Do a screenshot if you want to save the image.
;
; Guillermito
; guillermito@pipo.com
; http://www.guillermito2.net
; September 27th 2004
;
; Assemble it with:
;   TASM32 /ml /m3 /z /t Chi-Square
;   TLINK32 -Tpe -aa Chi-Square,,,import32 
;   UPX -9 Chi-Square.exe
;   Rebuild with LordPE (by yoda) to shrink it to the maximum.


callW macro   x
extrn x:PROC
call x
endm

   .486
   .model flat

;************************************************************
;************************************************************
;************************ DATA ******************************
;************************************************************
;************************************************************

.data

;------------------ structures --------------------

openfilename_struct:
lStructSize       dd openfilename_struct_size
hwndOwner         dd 0
hInstance         dd 0
lpstrFilter       dd offset filter
lpstrCustomFilter dd 0
nMaxCustFilter    dd 40
nFilterIndex      dd 0
lpstrFile         dd offset namebuffer
nMaxFile          dd 255
lpstrFileTitle    dd 0
nMaxFileTitle     dd 32
lpstrInitialDir   dd 0
lpstrTitle        dd 0
Flags             dd 1000h+4h+200000h
nFileOffset       dw 0
nFileExtension    dw 0
lpstrDefExt       dd 0
lCustData         dd 0
lpfnHook          dd 0
lpTemplateName    dd 0
openfilename_struct_size equ $-offset openfilename_struct
namebuffer            db 255 dup(0)
filter db "*.bmp",0, "*.bmp",0

wndclass:
clsStyle          dd 4003h
clsLpfnWndProc    dd ?
clsCbClsExtra     dd 0
clsCbWndExtra     dd 0
clsHInstance      dd ?
clsHIcon          dd 0
clsHCursor        dd 0
clsHbrBackground  dd 0
clsLpszMenuName   dd 0
clsLpszClassName  dd ?

msg:
msHWND          dd ?
msMESSAGE       dd ?
msWPARAM        dd ?
msLPARAM        dd ?
msTIME          dd ?
msPT            dd ?
zz              dd ?

;--------------- file/memory stuff ----------------------

nom_fenetre        db "Guillermito's Chi-Square Steganography Test",0
no_bmp             db "This file does not have the 24-bits BMP signature",0
handle_wd          dd ?
adresse_retour     dd ?

file_handle        dd ?
file_size          dd ?
file_mem_buffer    dd ?
file_nb_bytes_read dd ?
nb_chunks          dd ?
buf_tables         dd ? 
pixels_data_start  dd ?

;--------------- floats --------------------

float TYPEDEF REAL4
fchisquare float 0.0
fchisqprob float 0.0

normal_table:
 float  0.0
float   0.5000
 float  0.0627
float   0.4750
 float  0.1257
float   0.4500
 float  0.1891
float   0.4250
 float  0.2533
float   0.4000
 float  0.3186
float   0.3750
 float  0.3853
float   0.3500
 float  0.4538
float   0.3250
 float  0.5244
float   0.3000
 float  0.5978
float   0.2750
 float  0.6745
float   0.2500
 float  0.7554
float   0.2250
 float  0.8416
float   0.2000
 float  0.9346
float   0.1750
 float  1.0364
float   0.1500
 float  1.1503
float   0.1250
 float  1.2816
float   0.1000
 float  1.4395
float   0.0750
 float  1.6449
float   0.0500
 float  1.9600
float   0.0250
 float  2.3263
float   0.0100
 float  2.5758
float   0.0050
 float  3.0902
float   0.0010
 float  3.2905
float   0.0005
 float  3.7190
float   0.0001

ftemp float 0.0

;----------------- graph stuff ----------------------------

largeur   dd ?
hauteur   dd ?
buf_graph dd ?
counter_x dd ?
lsb0      dd 0
lsb1      dd 0
lsb_mean  dd ?
PoV_proba dd ?

chunk_size equ 128

;************************************************************
;************************************************************
;************************ CODE ******************************
;************************************************************
;************************************************************

.code

programme:

finit

;-------- choose a file -------------

push offset openfilename_struct
callW GetOpenFileNameA
test eax, eax
jz end

;-------- open it --------------

push 0
push 80h
push 3
push 0
push 0
push 80000000h+40000000h
push dword ptr [lpstrFile]
callW CreateFileA
inc eax
jz end
dec eax
mov file_handle, eax

;---------- get its size ---------

push 0
push dword ptr file_handle
callW GetFileSize
inc eax
jz close_file
dec eax
mov file_size, eax

;--------- alloc memory for file --------

mov eax, file_size
add eax, 1024
push eax
push 40h
callW LocalAlloc
test eax, eax
jz close_file
mov file_mem_buffer, eax

;------- read entire file ------------

push 0
push offset file_nb_bytes_read
push file_size
push file_mem_buffer
push file_handle
callW ReadFile
test eax, eax
jz free_memory
mov eax, file_nb_bytes_read
cmp eax, file_size
jnz free_memory

;-------- find BMP signature ---------

mov esi, file_mem_buffer
lodsw
cmp ax, "MB"
jne trouble

;------- find number of bits per pixel ------

add esi, 8
lodsd
add eax, file_mem_buffer
mov pixels_data_start, eax
add esi, 14
lodsw
cmp ax, 24
je this_looks_like_a_bmp

;-------- not 24-bits BMP -----------

trouble:
push 0
push offset nom_fenetre
push offset no_bmp
push 0
callW MessageBoxA
jmp free_memory

;---------- how many 128-bytes chunks ---------------------

this_looks_like_a_bmp:
mov eax, pixels_data_start
sub eax, file_mem_buffer
sub file_size, eax

fild file_size                 ;take file size
mov ftemp, chunk_size
fild ftemp
fdiv                           ;divide it by chunk size
fistp ftemp
mov ecx, ftemp
mov nb_chunks, ecx             ;save this number, needed to prepare the graph
                               ;(one chunk of X bytes will be one pixel wide)

   ;************************************************************
   ;*********** prepare bitmap for the graph *******************
   ;************************************************************
   
   pusha
   
   ;-------- take padding into account -------------
   
   mov eax, nb_chunks        ;this will be the width
   and eax, 11b              ;round it 
   mov ebx, 4                ;to the next 4 bytes / pixels
   sub ebx, eax
   mov eax, nb_chunks
   add eax, ebx
   add eax, 20               ;and add 20 more for some space
   
   ;-------- calculate needed memory --------------
   
   push eax
   pop largeur               ;eax becomes the width
   mov hauteur, 110          ;height is fixed at 110
   fild largeur
   fild hauteur
   fmul
   fistp buf_graph
   add buf_graph, 40+(256*4)+1024+((256*4)*2)+1024
   
      ;need room for:
      ;  bmp memory header             : 40
      ;  palette                       : 256*4
      ;  pixels                        : largeur*hauteur
      ;  table measured freq           : 256*4
      ;  table theorical freq          : 256*4
      ;  some buffer space between them: 1024*2

   push buf_graph               ;save for later            

   push buf_graph               ;number of bytes to allocate
   push 40h
   callW LocalAlloc
   test eax, eax
   jz free_memory

   ;---------- remember the pointers -----------------
      
   mov buf_graph, eax           ;bmp will start here
   mov edi, eax
   
   pop eax
   add eax, buf_graph
   sub eax, ((256*4)*2)+1024
   mov buf_tables, eax          ;tables will start here
   
   ;---- construct BITMAPINFOHEADER in memory --------------

      ;BITMAPINFO:
      ; BITMAPINFOHEADER: 
      ; biSize          dd bmpinfosize  ;biSize
      ; biWidth         dd largeur      ;biWidth
      ; biHeight        dd hauteur      ;biHeight
      ; biPlanes        dw 1            ;biPlanes (tjs 1)
      ; biBitCount      dw 8            ;biBitCount (8=256 couleurs)
      ; biCompression   dd 0            ;biCompression (BI_RGB=0, no compr)
      ; biSizeImage     dd 0            ;biSizeImage (0 si pas de compr)
      ; biXPelsPerMeter dd 0            ;biXPelsPerMeter
      ; biYPelsPerMeter dd 0            ;biYPelsPerMeter
      ; biClrUsed       dd 0            ;biClrUsed
      ; biClrImportant  dd 0            ;biClrImportant
      ; bmpinfosize equ $-offset BITMAPINFO (generally 40d)
      ;PALETTE db 256*4 dup (99h)               ;256 couleurs * (RGB + reserved) = 256 * 4 bytes

   xor eax, eax
   mov al, 40d
   stosd
   mov eax, largeur
   stosd
   mov eax, hauteur
   stosd
   mov eax, 00080001h
   stosd
   
   ;------------ fill palette (BGR0) --------------
   
   add edi, 6*4         ;go to palette start
   mov eax, 0           ;color 0 = black (background)
   stosd
   mov eax, 0ffffffh    ;color 1 = white (numbers)
   stosd
   mov eax, 0ff0000h    ;color 2 = red (chi-square)
   stosd
   mov eax, 00000ffh    ;color 3 = blue (axis)
   stosd
   mov eax, 0000055h    ;color 4 = dark blue (middle axis)
   stosd
   mov eax, 000ff00h    ;color 5 = green (average)
   stosd
   
   ;--------- draw Y axes -----------
   
   add edi, (256*4)-24  ;go to start of pixels
   push edi             ;save this pointer for later
   mov eax, largeur
   mov ecx, 5           ;starts 5 pixels from the bottom
   mul ecx
   add edi, eax
   add edi, 15          ;starts 15 pixels from the left side
   mov ecx, 100         ;go up for 100 pixels (0% to 100%)
   mov edx, nb_chunks   ;to draw the right side
   mov al, 3            ;color
   axe_y:
   mov [edi], al        ;left
   mov [edi+edx],al     ;right
      pusha
      inc al
      mov ecx, edx
      shr ecx, 6
      onekb_marker:
      add edi, 64 
      mov [edi], al     ;lines at 1024 bytes of LSB (if chunk_size is 128)
      loop onekb_marker
      popa
   add edi, largeur
   loop axe_y
   
   ;-------- draw X axis -------------
   
   pop edi
   push edi
   mov eax, largeur
   mov ecx, 5           ;starts 5 pixels from the bottom
   mul ecx
   add edi, eax
   add edi, 15          ;starts 15 pixels from the left
   mov eax, largeur
   mov ecx, 50          ;line at 0.5
   mul ecx
   mov edx, eax
   mov ebx, edx
   add ebx, edx         ;line on top
   mov ecx, nb_chunks   ;go right for this number of pixels
   inc ecx
   mov al, 3            ;color
   axe_x:
   mov [edi], al        ;bottom
      inc al
      mov [edi+edx],al  ;middle
      dec al
   mov [edi+ebx],al     ;top
   inc edi
   loop axe_x
   
   ;----------- draws a small 0 ----------------

      ; .***
      ; .*.*
      ; .*.*
      ; .*.*
      ; .***
   
   pop edi
   push edi
   mov eax, largeur
   mov ecx, 5             ;5 pixels from the bottom side
   mul ecx
   add edi, eax
   add edi, 5             ;10 pixels from the left side
   mov [edi], 01010100h
   add edi, largeur
   mov [edi], 01000100h
   add edi, largeur
   mov [edi], 01000100h
   add edi, largeur
   mov [edi], 01000100h
   add edi, largeur
   mov [edi], 01010100h

   ;----------- draws a small 1 ----------------

      ; ..*. 
      ; .**. 
      ; ..*. 
      ; ..*. 
      ; .***
   
   pop edi
   mov eax, largeur
   mov ecx, 100
   mul ecx
   add edi, eax
   add edi, 5
   mov [edi], 01010100h
   add edi, largeur
   mov [edi], 00010000h
   add edi, largeur
   mov [edi], 00010000h
   add edi, largeur
   mov [edi], 00010100h
   add edi, largeur
   mov [edi], 00010000h
   
   popa

   ;************************************************************
   ;******************** done with the graph *******************
   ;************************************************************

mov esi, pixels_data_start  ;starts at beginning of pixel data

process_whole_image:
push ecx                    ;save chunk counter

;---------- count Pairs of Values in the current chunk ---------------------

mov ecx, chunk_size          ;one chunk is X bytes
mov fchisquare, 0            ;the chi-square is initialized each time
mov edi, buf_tables          ;first (measured) table
mov lsb0, 0                  ;reset LSB mean counter
mov lsb1, 0                  ;reset LSB mean counter

count_PoV_in_1_chunk:
xor eax, eax
lodsb                        ;get the image byte
push eax                     ;save it for later LSB mean counting
shl eax, 2                   ;multiply it by 4 to get a table pointer
inc dword ptr [edi+eax]      ;increment the corresponding value in the table  
pop eax                      ;get the image byte again
and eax, 1                   ;keep just LSB
test eax, eax                ;is it zero?
jz increase_0                ;yes: don't count
inc lsb1                     ;no: increase LSB counter
jmp zozo
increase_0: 
inc lsb0
zozo:   
loop count_PoV_in_1_chunk

;----- calculate theorical Pairs of Value if they are randomly distributed ---------

mov ecx, 128                 ;256 values so 128 PoVs
mov edi, buf_tables          ;pointer to the measured PoVs

calculate_all:
fild dword ptr [edi]         ;get the first measured value
fild dword ptr [edi+4]       ;get the second measured value (same except LSB)
fadd                         ;add them
mov ftemp, 2
fild ftemp
fdiv                         ;divide by 2
add edi, 256*4               ;points to the corresponding values in the 2nd table
fstp dword ptr [edi]         ;put that value in the second (theorical) table
mov eax, dword ptr [edi]
mov dword ptr [edi+4], eax   ;and the next value is the same if they are random
sub edi, (256*4)-8           ;go back to the measured table, next pair of values
loop calculate_all

;----------------------- calculate chi-square -------------------------

mov ecx, 128                   ;128 pairs of values
mov edi, buf_tables            ;pointer to the measured first table
fld fchisquare                 ;load chi-square (equal to zero at this point)

calculate_chi_square:

fild dword ptr [edi]           ;load measured value
fld dword ptr [edi+(256*4)]    ;load corresponding theorical value
fsub                           ;substract them
fst ftemp
fld ftemp                      ;load the same value
fmul                           ;to multiply it by itself: square
cmp dword ptr [edi+(256*4)], 0 ;are we going to divide by zero?
je avoid_divide_by_zero
fld dword ptr [edi+(256*4)]    ;no: load corresponding theorical value
jmp ooo
avoid_divide_by_zero:          ;yes:
fld1                           ;load 1 instead
ooo:
fdiv                           ;divide
fadd                           ;add to the chi-square
add edi, 8                     ;go to the next pair of value
loop calculate_chi_square

fstp fchisquare                ;save final value of chi-square

;--------- chi-square to normal for n degrees of freedom ---------------

fld fchisquare

mov ftemp, 2
fild ftemp
fmul                           ;chi-square * 2
fsqrt                          ;square root of it
mov ftemp, (2*127)-1           ;128 degrees of freedom (I guess?)
fild ftemp
fsqrt                          ;square root of it
fsub                           ;substract both

fldz                           ;trick to know if number is negative: load zero
fcomp                          ;compare number to zero
fstsw ax                       ;store status word in ax
and eax, 100000000b            ;mask other bits
test eax, eax                  ;is this status bit at zero?
jz negative                    
mov edx, 1                     ;no: >0 positive number, use edx to remember
jmp jump_over
negative:
mov edx, -1                    ;yes: <0 negative number, use edx to remember
jump_over:

fabs                           ;get the absolute value
fstp fchisqprob

;----------- normal to probability -------------------------

push ecx
mov ecx, 25                    ;my normal table has 25 entries
mov edi, offset normal_table   ;points at the start
xor ebx, ebx                   ;will be used as table pointer
fld fchisqprob                 ;load normal value

go_through_table:
fld dword ptr [ebx+edi]        ;load current chi-square value from table
fcomp                          ;compare with our value
fstsw ax                       ;store status word in ax
and eax, 100000000b            ;mask other bits
test eax, eax                  ;is this status bit at zero?
jz break_loop 
                               ;no: our value is > current value in the table
add ebx, 8                     ;so point to next value in the table
loop go_through_table          ;and start again

break_loop:                    ;yes: we stop here
fstp ftemp                     ;empty the coprocessor stack
add ebx, 4                     ;point to the next dw which is the probability 
fld dword ptr [ebx+edi]        ;load it
cmp edx, 1                     ;were we positive some time ago?
je no_correction               ;yes: no need to modify
fld1                           ;no: the actual proba is 1-proba, load 1
fxch                           ;invert both values
fsub                           ;1-proba

no_correction:
mov ftemp, 100                 
fild ftemp
fmul                           ;multiply by 100 to get a percentage
fistp PoV_proba                ;unstack it in integer format

;------------------ calculate LSB average -----------------------

fild lsb1                      ;load how many LSBs at 1
mov ftemp, chunk_size          ;load total bit number
fild ftemp
fdiv                           ;divide
mov ftemp, 100                 ;multiply by 100 to fit in the graph
fild ftemp
fmul
fistp lsb_mean

   ;************************************************************
   ;********************* draw pixels **************************
   ;************************************************************

   ;-------------- chi-square red pixel -------------------
   
   pusha
   mov esi, buf_graph          ;pointer to start of bitmap memory
   add esi, 40+(256*4)         ;pointer to start of pixel values
   add esi, 16                 ;start 16 pixels from the left
   push esi                    ;save for later (&&&)
   mov eax, largeur
   mov ecx, PoV_proba          ;add 0 pixel if 0%, 100 pixels if 100%
   add ecx, 5                  ;start 5 pixels from the bottom
   mul ecx
   add esi, eax                ;final pixel offset
   mov eax, 2                  ;color
   mov ebx, counter_x          ;keep track of where we are on the X axis
   mov [esi+ebx], al           ;plot the pixel
   inc counter_x 
   
   ;----------- LSB average green pixel -------------------
   
   pop esi                     ;already calculated (&&&)
   mov eax, largeur
   mov ecx, lsb_mean           ;add 0 pixel if 0%, 100 pixels if 100%
   add ecx, 5                  ;start 5 pixels from the bottom
   mul ecx
   add esi, eax                ;final pixel offset
   mov eax, 5                  ;color
   mov ebx, counter_x          ;keep track of where we are on the X axis
   mov [esi+ebx], al           ;plot the pixel

   popa

;------------------ clean and start again ------------------

;pop ecx
pop ecx
pop ecx
dec ecx
jz breaque
jmp process_whole_image
breaque:

   ;******************************************************************************
   ;**********************  SHOW FINAL GRAPH  ************************************
   ;******************************************************************************

   ;--------- register wndclass ---------------------
   
   push 0
   callW GetModuleHandleA
   push eax                     ;@@ handle
   mov clsHInstance, eax
   mov eax, offset wndproc
   mov clsLpfnWndProc, eax
   mov clsLpszClassName, offset nom_fenetre
   push offset wndclass
   callW RegisterClassA
   
   ;------------- create window -----------------------
   
   pop ebx                      ;@@ handle
   xor eax, eax
   push eax
   push ebx                     ;handle
   push eax
   push eax
      mov ebx, hauteur
      add ebx, 27
   push ebx                     ;hauteur
      mov ebx, largeur
      add ebx, 8
   push ebx                     ;largeur
   push 400                     ;y
   push 10                      ;x
   push 214D0000h               ;normal window
   mov eax, offset nom_fenetre
   push eax
   push eax
   push 0
   callW CreateWindowExA
   mov handle_wd, eax
   
   push 1
   push handle_wd
   callW ShowWindow
   
   ;--------------- message loop ------------------
   
   msg_loop:
                
   xor eax, eax
   push 1
   push eax
   push eax
   push eax
   push offset msg
   callW PeekMessageA
   test eax, eax
   jnz process_messages

   ;------------ get DC of the window -----------------
   
   push handle_wd
   callW GetDC                  ;get device context
   push eax                     ;@@@@
   
   ;------- bitmap in memory to visible window ---------------
   
   xor ecx, ecx
   push 0CC0020h                ;SRCCOPY
   push ecx                     ;DIB_RGB_COLORS
   mov ebx, buf_graph           ;pointer to start of header
   push ebx
   add ebx, 40+(256*4)          ;pointer to start of pixels
   push ebx
   push hauteur                 ;hauteur DIB source
   push largeur                 ;largeur DIB source
   push ecx                     ;y du rectangle source
   push ecx                     ;x du rectangle source
   push hauteur                 ;hauteur rect destination
   push largeur                 ;largeur rect destination
   push ecx                     ;y du rectangle destination
   push ecx                     ;x du rectangle destination
   push eax                     ;DC of da window
   callW StretchDIBits          ;no stretch cause they are same size
   
   ;------- release DC, eax already pushed @@@@ -----------------
   
   push handle_wd
   callW ReleaseDC                      ;release  DC

   jmp msg_loop
   
;******************************************************************************
;**********************  WINDOW PROCEDURE  ************************************
;******************************************************************************

;-------------- see if window is closed -------------------

process_messages:
cmp msMESSAGE, 12h            ;WM_QUIT 
je free_memory
  
push offset msg
callW TranslateMessage
push offset msg
callW DispatchMessageA
jmp msg_loop

;------- close memory and file ----------------

free_memory:
push file_mem_buffer
callW LocalFree

push buf_graph
callW LocalFree

;----------- exit ----------

close_file:
push dword ptr file_handle
callW CloseHandle

end:
push -1
callW ExitProcess

;--------------- window proc ---------------------

wndproc:
pop adresse_retour
cmp dword ptr [esp+4], 2       ;WM_DESTROY?
jne end_wndproc                ;no => continue

push 0
callW PostQuitMessage          ;yes => quit
xor eax, eax

end_wndproc:
callW DefWindowProcA           ;call default proc with everything on the stack
push adresse_retour            ;put back return offset on stack
;mov eax, 2                     ;HTCAPTION so right click moves the window

ret

end programme
